/*
 * additive.cpp
 *   Two parameter model estimation helper functions
 *
 *  Created on: Feb 2, 2009
 *      Author: Apostolos
 */

#include <algorithm>
#include <vector>
#include <time.h>

#include "br2.h"
#include "utils.h"
#include "random_gen.h"

#include "plinkModule.h"

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;

/*
 *  Run br2 for the additive model with 2 parameters
 *  Unfortunately this is very similar to the logistic model, but we can't factor out the common code
 *  So this means, that for the time being, any change to the corresponding functions in br2.cpp should be
 *  propagated here.
 *
 */

extern plinkModule plinkRunner;

namespace Genotypic {

// Number of significant SNPs in Original dataset
int Sn;

// Un: Naives estimates
vector<GenotypicResults> Un;

/* mergeResults: Merge the genotypic results into the main results class */
void mergeResults(vector<GenotypicResults> & out, vector<GenotypicOutput> & geno)
{
    // plink output is sorted by CHR/BP so sort Un to be able to merge
    sort(out.begin(), out.end(), GenotypicResults_sortByLocus);
    sort(geno.begin(), geno.end(), GenotypicOutput_sortByLocus);

    for (uint i = 0; i < geno.size(); i++) {
	out[i].UdOR[0] = geno[i].ADD;
	out[i].UdOR[1] = geno[i].DOMDEV;
	out[i].UdA1 = geno[i].a1;

	out[i].adjustNaive();

	if (out[i].snp != geno[i].snp) {
	    cout
		    << "Error: mergeResults: cannot parse plink results, snp's are not in expected order.\n";
	    cout << "Expected: " << out[i].snp->name << " Got: " << geno[i].snp->name << "\n";
	    gracefulExit(1);
	}
    }

    sort(out.begin(), out.end(), GenotypicResults_sortByP);
}

void mergeResults(vector<GenotypicResults> & out, vector<GenotypicOutput> & Ud, vector<
	GenotypicOutput> &Ue)
{
    // plink output is sorted by CHR/BP so sort Un to be able to merge
    sort(out.begin(), out.end(), GenotypicResults_sortByLocus);
    sort(Ud.begin(), Ud.end(), GenotypicOutput_sortByLocus);
    sort(Ue.begin(), Ue.end(), GenotypicOutput_sortByLocus);

    if (Ud.size() != Ue.size()) {
	cout << "Error: mergeResults: didn't get the same # of snps for D and E steps.\n";
	gracefulExit(1);
    }

    for (uint i = 0; i < Ud.size(); i++) {
	out[i].UdOR[0] = Ud[i].ADD;
	out[i].UdOR[1] = Ud[i].DOMDEV;
	out[i].UdA1 = Ud[i].a1;

	out[i].UeOR[0] = Ue[i].ADD;
	out[i].UeOR[1] = Ue[i].DOMDEV;
	out[i].UeA1 = Ue[i].a1;

	out[i].adjust();

	//printf("MERGE: %10s Ud ADD %5g DOM %5g Ue ADD %5g DOM %5g\n", out[i].snp->name.c_str(),
	//		out[i].ADD_UdOR, out[i].DOM_UdOR, out[i].ADD_UeOR, out[i].DOM_UeOR);

	if (out[i].snp != Ud[i].snp || out[i].snp != Ue[i].snp) {
	    cout
		    << "Error: mergeResults: cannot parse plink results, snp's are not in expected order.\n";
	    cout << "Expected: " << out[i].snp->name << " Got: " << Ud[i].snp->name << " "
		    << Ue[i].snp->name << "\n";
	    gracefulExit(1);
	}
    }

    sort(out.begin(), out.end(), GenotypicResults_sortByP);
}

vector<Locus*> getNaiveEstimate()
{
    char s[1024]; // temporary storage for printing

    vector<ModelOutput> UnSig;
    string baseOutFile = tmpdir + "temp_all";

    cout << "Computing naive estimate...\n\n";

    // Check significant SNPs in original dataset
    Sn = isSignificant(baseOutFile, persons, locus, 0, UnSig);

    if (Sn == 0) {
	cout << "\nNo Significant SNPs in input file. Nothing to do. \n";
	gracefulExit(1);
    }

    sort(UnSig.begin(), UnSig.end()); // sort by p-value

    for (int i = 0; i < UnSig.size(); i++)
	UnSig[i].snp->rank = i + 1;

    sprintf(s, "Original dataset: # of significant snps = %d \n", Sn);
    cout << s;

    // Select only top <k> SNPs for analysis
    if (maxSNPs != -1) { // option --top specified
	cout << "Selecting only top " << maxSNPs << " significant SNPs\n";

	int sig = 0;
	for (uint i = 0; i < UnSig.size(); i++)
	    if (UnSig[i].significant) {
		if (sig + 1 > maxSNPs) {
		    UnSig[i].significant = 0;
		    continue;
		}
		sig++;
	    }

	Sn = sig;
    }

    // Keep only top <k> significant SNPs for bootstrapping and delete all others
    if (keeptop != -1) { // option --keep specified
	cout << "Option --only specified: Keeping only " << keeptop << " SNPs\n";

	if (keeptop < Sn) {
	    cout << "Error: Cannot keep less than the number of significant SNPs\n";
	    gracefulExit(1);
	}

	for (uint i = keeptop; i < UnSig.size(); i++)
	    UnSig[i].snp->id = -1;
	deleteSNPS();
	cout << "Locus size now is " << locus.size() << "\n";

	// Now run function again
	keeptop = -1;
	return getNaiveEstimate();
    }

    if (silent == 0) {
	sprintf(s, "Significant SNPs:\n");
	cout << s;

	for (uint i = 0; i < UnSig.size(); i++)
	    if (UnSig[i].significant) {
		sprintf(s, "%s(%g) ", UnSig[i].snp->name.c_str(), UnSig[i].p);
		cout << s;
	    }

	sprintf(s, "\n\n");
	cout << s;
    }

    // Select only signigicant SNPs for analysis
    vector<Locus*> selected;
    for (unsigned int i = 0; i < UnSig.size(); i++)
	if (UnSig[i].significant) {
	    selected.push_back(UnSig[i].snp);
	    GenotypicResults r;

	    r.snp = UnSig[i].snp;
	    r.p = UnSig[i].p;
	    r.a1 = 'x';

	    Un.push_back(r);
	}

    return selected;
}

/*
 * Function getNaiveOR:
 *  run plink and get original Un estimate;
 *
 *  persons: Which persons to use
 *  selectedSNPs: Which SNPs to use
 *
 */

void getNaiveOR(vector<Individual*> & persons, vector<Locus*> &selectedSNPS)
{
    vector<GenotypicOutput> UnLog;
    char s[100];

    // plink will output these sorted, so we'd better sort them now otherwise reading will be harder
    stable_sort(selectedSNPS.begin(), selectedSNPS.end(), Locus_sortByCHR);
    plinkRunner.writeBinFiles(tmpdir + "temp_un", persons, selectedSNPS, 0);
    plinkRunner.genotypic(tmpdir + "temp_un", tmpdir + "temp_un", UnLog, selectedSNPS);

    // Merge plink results into main matrix Un
    mergeResults(Un, UnLog);

    if (silent == 0) {
	cout << "Naive p-values and OR:\n";
	for (uint i = 0; i < Un.size(); i++) {
	    sprintf(s, "%s %g %g %g (%c)\n", Un[i].snp->name.c_str(), Un[i].p, pow(
		    M_E, Un[i].UdOR[0]), pow(M_E, Un[i].UdOR[1]), Un[i].UdA1);
	    cout << s;
	}
	cout << "\n\n";
    }

    cout.flush();
}

/*
 * getBias: computes Ud_Ue component of bias for this bootstrap sample
 *
 */
int getBias(vector<GenotypicResults> & out, BootStrap &BS, vector<Locus*> & SNPs,
	int maxSNPsToProcess, char *id, char *thread_id)
{
    char s[1000];

    string prefix = tmpdir + "temp_" + thread_id;

    vector<ModelOutput> model;

    // Check if bootstrap sample is significant
    int numSig = isSignificant(prefix + "all", BS.BootstrapSample, SNPs, 1, model);

    sprintf(s, "Level %s:  significant snps=%d \n", id, numSig);
    cout << s;

    //outputPvalues((string)"p" + (string)id, model);

    out.clear();

    if (numSig == 0) return 0;

    // Sample is significant, compute Ud/Ue quantities

    // Sort model results by p value
    sort(model.begin(), model.end());

    if (debug) {
#pragma omp critical
	{
	    for (int i = 0; i < model.size(); i++)
		model[i].snp -> ranks.push_back(i + 1);

	}
    }

    /*
     *  Select only signigicant SNPs for analysis
     */

    vector<Locus*> selected;
    out.reserve(numSig);

    int sig = 0;

    for (unsigned int i = 0; i < model.size(); i++)
	if (model[i].significant) {
	    selected.push_back(model[i].snp);
	    GenotypicResults r(model[i].snp, model[i].p);
	    out.push_back(r);
	    sig++;

	    // Break if we exceed maxSNPsToProcess
	    if (maxSNPsToProcess != -1 && sig >= maxSNPsToProcess) break;
	}

    /* Sort by CHR/BP */

    stable_sort(selected.begin(), selected.end(), Locus_sortByCHR);

    /* Compute Ud/Ue */

    vector<GenotypicOutput> Ud, Ue;

    string udfile = prefix + "ud";
    string uefile = prefix + "ue";

    plinkRunner.writeBinFiles(udfile, BS.BootstrapSample, selected, 1);
    plinkRunner.writeBinFiles(uefile, BS.NotSelected, selected, 1);

    plinkRunner.genotypic(udfile, udfile, Ud, selected);
    plinkRunner.genotypic(uefile, uefile, Ue, selected);

    mergeResults(out, Ud, Ue);

    return out.size();
}

void print_bias(const string &label, const string &prefix, vector<GenotypicResults> &v)
{
    unsigned int n = v.size();
    char s[1000];
    if (silent == 1) return;
    cout << "Matrix " << label << "\n";

    for (unsigned int i = 0; i < n; i++) {
	sprintf(
		s,
		"%s %15s p=%12g\tADD UdOR,A1=(%4.2f,%c)\tUeOR,A1=(%4.2f,%c) DOMDEV UdOR,A1=(%4.2f,%c)\tUeOR,A1=(%4.2f,%c)\n",
		prefix.c_str(), v[i].snp->name.c_str(), v[i].p, pow(M_E, v[i].UdOR[0]), v[i].UdA1,
		pow(M_E, v[i].UeOR[0]), v[i].UeA1, pow(M_E, v[i].UdOR[1]), v[i].UdA1, pow(
			M_E, v[i].UeOR[1]), v[i].UeA1);
	cout << s;
    }
}

vector<double> Uboot0ADD; // Bias reduced OR estimate, computed from level 0 quantities
vector<double> Uboot0DOMDEV; // Bias reduced OR estimate, computed from level 0 quantities

void getLevel0Quantities()
{
    vector<int> S0;
    vector<vector<GenotypicResults> > L0;

    int N = 0;
    int m = 0;
    char s[500];

    cout << "Computing Level 0 quantities.\n\n";
    //genAllPheno();

    // used to compute estimated time left
    time_t start = time(NULL);

    L0.clear();
    L0.reserve(Level0Bootstraps + 10);

    S0.clear();
    S0.reserve(Level0Bootstraps + 10);

    int iteration = 0;

    // The max number of iterations is 10*specified number, to avoid running indefinetely
    // in case we can't get significant bootstraps

    // loop has been parallelized using OpenMP

    // It is not possible to break from a loop any time when using openmp
    // So we run 10*bootstraps and when we get enough significant ones
    // we just do continue until loop is finished

#pragma omp parallel for shared(m)
    for (N = 0; N < 10 * Level0Bootstraps; N++) {
	if (m >= Level0Bootstraps) continue; // skip if we got enough bootstraps

	char id[100] = "", thread_id[100] = "";
	int myIteration;

	BootStrap L0BS(&persons);

	// This has to be critical since the random number generator is not re-entrant

#pragma omp critical
	{
	    L0BS.generateBootstrap();
	    myIteration = iteration++; // get an iteration id
	}

#ifdef _OPENMP
	if(silent == 0) printf("Thread: %d N=%d m=%d L0=%d\n",(int)omp_get_thread_num(), N, m, (int)L0.size());
	sprintf(thread_id,"%d", (int)omp_get_thread_num());
#endif

	sprintf(id, "L0-%d", myIteration);

	vector<GenotypicResults> out;
	int numSig = getBias(out, L0BS, locus, Sn, id, thread_id);
	if (numSig == 0) continue;

	// Add entries to S0 and L0
#pragma omp critical
	{
	    L0.push_back(out);
	    S0.push_back(numSig);
	    if (numSig >= Sn) m++;
	    //if(m >= Level0Bootstraps) break;

	    print_bias("L0[N]", "  ", out);
	    cout.flush();
	}

	if (0 && m == 2) {
	    char s[500];
	    time_t now = time(NULL);
	    double time_est = Level0Bootstraps * ((double) now - start + 1) / m;

	    if (time_est > 100) sprintf(s, "%.1f min", time_est / 60.0);
	    else sprintf(s, "%f sec", time_est);

	    cerr << "Estimated time to finish L0: " << s << " (" << ((int) (59 + time_est - (now
		    - start)) / 60) << " min left)\n";

	}

    }

    N = L0.size();
    cout << "Level 0 completed. Significant Bootstraps were " << ((double) m / N) * 100 << "% \n";

    printf("Sn=%d\n S0=", (int) Sn);
    for (int i = 0; i < S0.size(); i++)
	printf("%d ", (int) S0[i]);
    printf("\n");
    cout << L0.size() << " " << S0.size() << "\n";
    cout << N << " " << m << "\n";

    /**
     *   Compute Uboot0[k] k=1..Sn
     */

    for (int k = 0; k < Sn; k++) { // For each significant snp

	int ADD_total = 0;
	double ADD_avg = 0;

	for (int b = 0; b < N; b++) {
	    if (L0[b].size() > k && L0[b][k].UdOR[0] != NA && L0[b][k].UeOR[0] != NA) {

		ADD_avg += L0[b][k].UdOR[0] - L0[b][k].UeOR[0];
		ADD_total++;

	    }
	}

	if (ADD_total != 0) Uboot0ADD.push_back(Un[k].UdOR[0] - ADD_avg / ADD_total);
	else Uboot0ADD.push_back(NA);

	int DOM_total = 0;
	double DOM_avg = 0;

	for (int b = 0; b < N; b++) {
	    if (L0[b].size() > k && L0[b][k].UdOR[1] != NA && L0[b][k].UeOR[1] != NA) {

		DOM_avg += L0[b][k].UdOR[1] - L0[b][k].UeOR[1];
		DOM_total++;

	    }
	}

	if (DOM_total != 0) Uboot0DOMDEV.push_back(Un[k].UdOR[1] - DOM_avg / DOM_total);
	else Uboot0DOMDEV.push_back(NA);

	sprintf(
		s,
		" Uboot0 for SNP %d:%s (%d samples used) ADD: Un=%g Uboot=%g DOMDEV: Un=%g Uboot=%g\n",
		k, Un[k].snp->name.c_str(), ADD_total, Un[k].UdOR[0], Uboot0ADD[k], Un[k].UdOR[1],
		Uboot0DOMDEV[k]);

	cout << s;
	cout.flush();
    }

    // Implement truncation of Uboot at 0
    //for (int i = 0; i < Sn; i++) if(Uboot0ADD[i] < 0) Uboot0ADD[i] = 0;
    //for (int i = 0; i < Sn; i++) if(Uboot0DOMDEV[i] < 0) Uboot0DOMDEV[i] = 0;
}

void outputFinalResults()
{
    char s[1000], CI_str[1000];

    vector<string> d;
    vector<vector<string> > data;

    cout << "\n\nAnalysis finished.\n";
    cout << "There were " << Sn << " significant SNPs in original dataset.\n";

    ofstream out;
    out.open(resultsFile.c_str(), ios::out);

    d.resize(100);

    int p = 0;

    d[p++] = "CHR";
    d[p++] = "   SNP";
    d[p++] = " BP";
    //d[p++] = " MAF";

    //d[p++] = " MAFCONT";
    //d[p++] = " MAFCASE";
    //d[p++] = "A1";
    //d[p++] = "A2";

    d[p++] = "        P";
    d[p++] = "G1_UN";
    d[p++] = "G1_UBOOT";
    d[p++] = "G2_UN";
    d[p++] = "G2_UBOOT";

    //d[p++] = "SD";
    //d[p++] = "CI";
    //d[p++] = "  RI";

    d.resize(p);
    data.push_back(d);

    for (int i = 0; i < Sn; i++) {

	/*if(StdDev[i] == NA)
	 sprintf(CI_str, "(%s,%s)", "NA","NA");
	 else
	 sprintf(CI_str, "(%.3f,%.3f)",
	 Uboot0[i] - 1.96 * StdDev[i],
	 Uboot0[i] + 1.96 * StdDev[i]);
	 */
	p = 0;

	d[p++] = toString(Un[i].snp->chr);
	d[p++] = Un[i].snp->name;
	d[p++] = toString(Un[i].snp->bp);
	//d[p++] = toString(Un[i].snp->maf,"%.3f");

	//d[p++] = toString(Un[i].snp->mafControls,"%.3f");
	//d[p++] = toString(Un[i].snp->mafCases,"%.3f");
	//d[p++] = Un[i].a1;
	//d[p++] = Un[i].snp->allele2;

	d[p++] = toString(Un[i].p, "%.3g");

#define exp

	d[p++] = toString(exp(Un[i].UdOR[0]), "%.3f");
	d[p++] = toString(exp(Uboot0ADD[i]), "%.3f");

	d[p++] = toString(exp(Un[i].UdOR[1]), "%.3f");
	d[p++] = toString(exp(Uboot0DOMDEV[i]), "%.3f");

	//d[p++] = toString(StdDev[i],"%.3f");
	//d[p++] = CI_str;
	//d[p++] = toString(RI[i],"%g");

	data.push_back(d);
    }

    printTable(cout, data);
    printTable(out, data);
}

/*
 * runBootstrap(): main function to perform double level bootstrap
 */

void runBootStrap()
{
    vector<Locus*> selected;

    selected = Genotypic::getNaiveEstimate();

    // get naive OR estimate
    Genotypic::getNaiveOR(persons, selected);

    Genotypic::getLevel0Quantities();
    //getLevel12Quantities();
    //Genotypic::computeFinalResults();
    Genotypic::outputFinalResults();

    return;
}

}
; // end namespace Genotypic


