/*
 * File:   plinkModule.cpp
 * Author: Apostolos Dimitromanolakis
 *
 * Created on October 17, 2008, 11:03 PM
 */


#include "plinkModule.h"
#include "readline.h"

#include <sstream>


extern int nlocus;
extern vector<Locus*> locus;

extern int npersons;
extern vector<Individual*> persons;

// Used to speed up file reading (see readline.h)
FILE *in;
int curline;

/*
 *  Initialize with default data
 */

void plinkModule::init()
{
    plinkLocation = "plink";
    plinkOptions = "--allow-no-sex --noweb --silent";
}

/*
 * Deallocate items from memory that no destructor frees
 */

plinkModule::~plinkModule()
{
}

/**
 *  Execute plink command line and check for errors
 *
 */

void plinkModule::runPLink(string & args)
{
    string cmdline = plinkLocation + " " + plinkOptions + " " + args + " ";

    if (debug >= 2) cout << "Executing [" << cmdline << "]\n";

    int ret = system(cmdline.c_str());

    if (ret != 0) {
	cout << "Error: execution of plink failed with exit code " << ret << "\n";
	cout << "  command was '" << cmdline << "'\n";
	gracefulExit(1);
    }
}

void plinkModule::addPlinkOption(const string &h)
{
    plinkOptions += " ";
    plinkOptions += h;
}

/*
 *  double plinkModule::checkPlinkVersion
 *     Check the plink version in use
 *
 *     Options:
 *       scratchDir  <->  Temporary directory to use
 *
 *     Returns:
 *       plink version as a double (for example 1.04)
 */

double plinkModule::checkPlinkVersion(string scratchDir)
{
    string logFile = scratchDir + "plinktest.log";
    string mapFile = scratchDir + "plinktest.map";
    string pedFile = scratchDir + "plinktest.ped";

    remove(logFile.c_str());

    ofstream map, ped;
    map.open(mapFile.c_str());
    if (!map) {
	cout << "Error: Cannot write to scratch directory: '" << scratchDir << "'\n";
	gracefulExit(1);
    }

    map << "1 snp 0 0\n";
    map.close();

    ped.open(pedFile.c_str());
    ped << "per0 per0 0 0 2 2 B A\n";
    ped.close();

    string cmdline = "--file " + scratchDir + "plinktest" + " --out " + scratchDir + "plinktest";
    runPLink(cmdline);

    remove(mapFile.c_str());
    remove(pedFile.c_str());

    ifstream in;
    in.open(logFile.c_str(), ios::in);
    if (!in) {
	remove(logFile.c_str());
	remove(mapFile.c_str());
	remove(pedFile.c_str());
	return -1;
    }

    double ver = -1;
    string s;
    while (in >> s) {
	if (s[0] == 'v') {
	    s[0] = '0';
	    ver = atof(s.c_str());
	    break;
	}
    }

    remove(logFile.c_str());
    remove(mapFile.c_str());
    remove(pedFile.c_str());

    return ver;
}

/*
 * void plinkModule::Logistic
 *   Run plink --logistic test on file binfile and read results file
 *
 */


void plinkModule::logistic(const string & binfile,
	vector<LogisticOutput> &results, vector<Locus*> &locus)
{
    // Run Plink

    string cmdline = "--bfile \"" + binfile + "\" " + "--logistic " + "--out \"" + binfile + "\"";

    if (haveCovariates == true) {
	cmdline += " --covar " + binfile + ".cov " + par::covariate_command_line;
    }

    if(locus.size() == 0) {
	results.clear();
	return;
    }

    runPLink(cmdline);

    // Read plink output file

    ifstream in;
    in.open((binfile + ".assoc.logistic").c_str());

    //system(("plink --bfile " + binfile + " --model --silent").c_str());
    //system(((string)"cat plink.model").c_str());
    //system(("cat " + binfile + ".assoc.logistic").c_str());

    if (!in) {
	cout << "Cannot open plink output file: " + (binfile + ".assoc.logistic") + "\n";
	gracefulExit(1);
    }

    results.clear();
    int nlocus = locus.size();

    char s[10000]; // Line currently read from input file

    int chr, bp;
    char snp[512], allele[16], test[32], nmiss_s[100], odds_s[100], stat_s[100], p_s[100];

    double odds, p;

    in.getline(s, 10000); // skip header
    int pos = 0;
    int line = 1;

    while (in.getline(s, 10000)) {
	line++;

	int n = sscanf(s, "%d %s %d %s %s %s %s %s %s", &chr, snp, &bp, allele, test, nmiss_s,
		odds_s, stat_s, p_s);

	if (n != 9) {
	    cout << "Error: Cannot read enough items from output file " << (binfile
		    + ".assoc.logistic") << " line " << line << "\n";
	    cout << "Offending line is: " << s << "\n\n";
	    gracefulExit(1);
	}

	// Skip covariates
	if(strcmp(test,"ADD") != 0) continue;

	// printf("%s %s %s %s\n",snp,test,odds_s,p_s);

	if (pos >= nlocus) {
	    cout << "Error: Additional SNPs in output file: " << (binfile + ".assoc.logistic")
		    << "!\n";
	    gracefulExit(1);
	}

	if (strcmp(locus[pos]->name.c_str(), snp) != 0) {
	    cout << "Error: SNP's out of order on output file " << (binfile + ".assoc.logistic")
		    << "    line:" << pos << "  expected:" << locus[pos]->name << "  got: " << snp
		    << "\n";

	    cout << "br2 input files should be sorted according to chr/bp. "
	            "Please recode input file with plink (plink --recode) and run br2 again.\n";

	    gracefulExit(1);
	}

	odds = toDouble(odds_s);
	p = toDouble(p_s);

	LogisticOutput r(locus[pos], allele[0], odds, p);
	results.push_back(r);

	pos++;
    }

    if(results.size() != locus.size()) {
	cout << "Error: Cannot read all logistic test results from plink (got "
	     << results.size() << " expected " << locus.size() << " )\n";

	gracefulExit(1);
    }
}

/*
 * void plinkModule::Linear
 *   Run plink --linear test on file binfile and read results file
 *
 */
void plinkModule::linear(const string & binfile,
	vector<LogisticOutput> &results, vector<Locus*> &locus)
{
    const string outfile = binfile;

    // Run Plink
    string cmdline = "--bfile \"" + binfile + "\" " + "--linear " + "--out \"" + outfile + "\"";
    string resultsFile = (outfile + ".assoc.linear");

    if (haveCovariates == true) {
	cmdline += " --covar " + binfile + ".cov " + par::covariate_command_line;
    }

    runPLink(cmdline);

    // Read plink output file

    ifstream in;
    in.open(resultsFile.c_str());

    if (!in) {
	cout << "Cannot open plink output file: " + resultsFile + "\n";
	gracefulExit(1);
    }

    results.clear();
    int nlocus = locus.size();

    char s[10000]; // Line currently read from input file

    int chr, bp;
    char snp[512], allele[16], test[32], nmiss_s[100], odds_s[100], stat_s[100], p_s[100];

    double odds, p;

    in.getline(s, 10000); // skip header
    int pos = 0;
    int line = 1;

    while (in.getline(s, 10000)) {
	line++;

	int n = sscanf(s, "%d %s %d %s %s %s %s %s %s", &chr, snp, &bp, allele, test, nmiss_s,
		odds_s, stat_s, p_s);

	if (n != 9) {
	    cout << "Error: Cannot read enough items from output file " << resultsFile
		<< " line " << line << "\n";
	    cout << "Offending line is: " << s << "\n\n";
	    gracefulExit(1);
	}

	// Skip covariates
	if(strcmp(test,"ADD") != 0) continue;

	// printf("%s %s %s %s\n",snp,test,odds_s,p_s);

	if (pos >= nlocus) {
	    cout << "Error: Additional SNPs in output file: " << resultsFile << "!\n";
	    gracefulExit(1);
	}

	if (strcmp(locus[pos]->name.c_str(), snp) != 0) {
	    cout << "Error: SNP's out of order on output file " << resultsFile
		    << "    line:" << pos << "  expected:" << locus[pos]->name << "  got: " << snp
		    << "\n";

	    cout << "br2 input files should be sorted according to chr/bp. "
	            "Please recode input file with plink (plink --recode) and run br2 again.\n";

	    gracefulExit(1);
	}

	odds = toDouble(odds_s);
	p = toDouble(p_s);

	LogisticOutput r(locus[pos], allele[0], pow(M_E,odds), p);
	if(odds == NA) r.odds = NA;

	// cout << locus[pos]->name << " " << allele[0] << " " << odds << " " << p << "\n";

	results.push_back(r);

	pos++;
    }

    if(results.size() != locus.size()) {
	cout << "Error: Cannot read all logistic test results from plink (got "
	     << results.size() << " expected " << locus.size() << " )\n";

	gracefulExit(1);
    }
}


/*
 * Model: Run plink --model and parse output file
 */
void plinkModule::model(const string & binfile, const string & outfile, vector<ModelOutput> &out)
{
    out.clear();

    // Run Plink
    string cmdline = "--bfile \"" + binfile + "\" " + "--model " + "--out \"" + outfile + "\" ";
    runPLink(cmdline);

    // Read plink output file
    string file = (outfile + ".model");

    ifstream in;
    in.open(file.c_str());

    if (!in) {
	cout << "Error: Cannot open plink output file: " + file + "\n";
	gracefulExit(1);
    }

    char s[10000];

    int chr;
    char snp[512], test[32], a1[10], a2[10], aff[100], unaff[100], chisq_s[100], df_s[100],
	    p_s[100];
    double chisq, df, p;

    in.getline(s, 10000); // skip header (first line)

    int pos = -1, line = 0;

    char lastSnp[512] = {
    0
    };

    ModelOutput m;
    m.a1 = 0;

    while (in.getline(s, 10000)) {
	line++;

	chisq = -1;
	df = -1;
	p = -1;

	int n = sscanf(s, "%d %s %s %s %s %s %s %s %s %s", &chr, snp, a1, a2, test, aff, unaff,
		chisq_s, df_s, p_s);

	if (n != 10) {
	    cout << "Error: Cannot read enough items from model file " << (outfile + ".model")
		    << " line " << line << "\n";
	    cout << "Offending line is: " << s << "\n\n";
	    gracefulExit(1);
	}

	chisq = toDouble(chisq_s);
	df = toDouble(df_s);
	p = toDouble(p_s);

	// Increment snp position if we get a different snp ID
	if (strcmp(snp, lastSnp) != 0) {

	    if (m.a1 != 0) {
		//cout << m.snp->name << " " << m.pTrend << "\n";
		out.push_back(m);
	    }

	    m.a1 = 0;
	    pos++;

	    // if some snps are ommited from output file based on filtering then advance locus pointer
	    // until we get a match

	    /*while (strcmp(locus[pos]->name.c_str(), snp) != 0) {
	     pos++;

	     if (pos >= nlocus) break;
	     }*/

	    if (pos >= nlocus) {
		cout << "Error: SNP's out of order on output file " << file << " line " << line
			<< " expected snp " << snp << "\n";

		cout
			<< "br2 input files should be sorted according to chr/bp. Please recode input file with plink (plink --recode) and run br2 again. \n";

		gracefulExit(1);
	    }

	    m.snp = locus[pos];
	    strcpy(lastSnp, snp);
	}

	if (strcmp(m.snp->name.c_str(), snp) != 0) {
	    cout << "Error: SNP's out of order on output file " << file << ", line " << pos << " "
		    << locus[pos]->name << "\n";

	    cout
		    << "br2 input files should be sorted according to chr/bp. Please recode input file with plink (plink --recode) and run br2 again. \n";
	    gracefulExit(1);
	}

	if (pos >= nlocus) {
	    cout << "Error: Extra SNPs in output file!\n";
	    gracefulExit(1);
	}

	m.a1 = a1[0];

	if (strcmp(test, "GENO") == 0) m.pGeno = p;
	else if (strcmp(test, "TREND") == 0) m.pTrend = p;

	//cout << locus[pos].name << "Read test:" << test << " pGeno=" << locus[pos].pGeno << " " << locus[pos].pTrend << "\n";
	//cout << snp << " " << stat << " " << p << " " << "\n";
	//pos++;
    }

    if (m.a1 != 0) out.push_back(m);

    if ((int) out.size() != nlocus) {
	cout << "Warning:   " << (nlocus - out.size()) << " SNPs were filtered based on maf.\n";

	//cout << "Error: Did not get enough model data from " + file
	//		+ " (only got data for " << out.size() << " SNPs)\n";
	//exit(1);
    }

}

/**
 * QTLAssoc: run association tests on quantitative trait data.
 *
 */
void plinkModule::QTLAssoc(const string& binfile, vector<ModelOutput> &out, vector<Locus*> &locus)
{
    writeBinFiles(binfile, persons, locus, 1);

    string cmdline = "--bfile \"" + binfile + "\" " + "--assoc " + "--out \"" + binfile + "\" ";
    runPLink(cmdline);
    // Not used, all computations are performed internally now
}

/**
 * readMapFile: Reading of a plink Map file
 *
 */

void plinkModule::readMapFile(const string & mapfile)
{
    ifstream map;
    map.open(mapfile.c_str());

    if (!map) {
	cout << "Error: Cannot open input file " + mapfile << "\n";
	gracefulExit(1);
    }

    int pos = locus.size();

    while (true) {
	Locus *a = new Locus();
	map >> a->chr >> a->name >> a->pos >> a->bp;

	if (!map) { delete a; break; } // end of file

	a->id = pos++;
	locus.push_back(a);
    }

    nlocus = locus.size();

    cout << "Map file contains " << nlocus << " SNP ids. \n";

    map.close();
}

/*
 *  Reading of a text ped file
 */

void plinkModule::readPedFile(const string & pedfile)
{
    in = fopen(pedfile.c_str(), "rb");

    if (!in) {
	cout << "Error: Cannot open input file " + pedfile << "\n";
	gracefulExit(1);
    }

    char iid[100], fid[100], pat[100], mat[100], sex[100], phenotype[100];
    int items;

    // Major and minor alleles, used to speed up file reading
    char *allele1, *allele2;

    allele1 = new char[nlocus];
    allele2 = new char[nlocus];
    for (int i = 0; i < nlocus; i++) {
	allele1[i] = 0;
	allele2[i] = 0;
    }

    curline = 1;

    initLine();

    while (readLine() >= 0) {
	int pp = 0;
	items = sscanf(line, "%s %s %s %s %s %s %n", fid, iid, pat, mat, sex, phenotype, &pp);
	line_pos += pp;

	if (items < 6) {
	    cout << "Error: Cannot read all items from ped file input line " << curline << "\n";
	    gracefulExit(1);
	}

	Individual *i = new Individual;

	i->iid = iid;
	i->fid = fid;
	i->pat = pat;
	i->mat = mat;
	i->sexcode = sex;

        i->phenotype = getPhenotype((string)phenotype);

	// Add necessary locus space
	i->resizeGeno(nlocus);

	//i->one.resize(nlocus);
	//i->two.resize(nlocus);

	char one, two; // the two alleles currently read
	int snp = 0; // number of SNP currently being read

	while (1) {

	    // Get the two alleles from ped file

	    one = nextGenoType();
	    if (one == -1 || one == EOF) break;

	    two = nextGenoType();
	    if (two == -1 || two == EOF) break;

	    if (snp >= nlocus) {
		printf("Error: Too many genotype data on line %d!\n", curline);
		gracefulExit(1);
	    }

	    if (one == '0' || two == '0') {
		i->setGeno(snp, 1);
		//i->one[snp] = true;
		//i->two[snp] = false;
		snp++;
		continue;
		// Geno is missing will handle later, don't worry
	    }

	    if (one != allele1[snp] && one != allele2[snp]) {
		if (allele1[snp] == 0) allele1[snp] = one;
		else if (allele2[snp] == 0) allele2[snp] = one;
		else printf("!! Error in genotype line %d, SNP %d : %c (%c %c)\n", curline, snp,
			one, allele1[snp], allele2[snp]);
	    }

	    if (two != one) if (two != allele1[snp] && two != allele2[snp]) {
		if (allele1[snp] == 0) allele1[snp] = two;
		else if (allele2[snp] == 0) allele2[snp] = two;
		else printf("!! Error in genotype line %d, SNP %d : %c (%c %c)\n", curline, snp,
			two, allele1[snp], allele2[snp]);
	    }

	    // 00 hom
	    if (one == allele1[snp] && two == allele1[snp]) {
		i->setGeno(snp, 0);
		//i->one[snp] = false;
		//i->two[snp] = false;
	    }// 01 het
	    else if (one != two) {
		i->setGeno(snp, 2);
		//i->one[snp] = false;
		//i->two[snp] = true;
	    }// 11 hom
	    else if (one == allele2[snp] && two == allele2[snp]) {
		i->setGeno(snp, 3);
		//i->one[snp] = true;
		//i->two[snp] = true;
	    }// 10 missing
	    else {
		i->setGeno(snp, 1);
		//i->one[snp] = true;
		//i->two[snp] = false;
	    }

	    // printf("%c%c ", one,two);

	    snp++;
	}

	if (snp != nlocus) {
	    printf("Error: Not enough SNP's read from ped file file %s line %d\n", pedfile.c_str(),
		    curline - 1);
	    gracefulExit(1);
	}

	i->id = persons.size();
	persons.push_back(i);

    }

    npersons = persons.size();

    /* Copy back locus allele data */
    for (int i = 0; i < nlocus; i++) {
	locus[i]->allele1 = allele1[i];
	locus[i]->allele2 = allele2[i];
    }

    delete[] allele1;
    delete[] allele2;

    cout << "Pedigree file contains " << npersons << " individuals.\n";

    freeLine();
    fclose(in);
}

/**
 *    Reading of text based ped/map files
 */

void plinkModule::readTextFiles(const string & pedfile, const string & mapfile)
{
    readMapFile(mapfile);
    readPedFile(pedfile);
}

/**
 *  Functions to write plink Binary Files
 */

/**
 * Write a .bim file
 * .bim files contain similar information as text .map files
 */

void plinkModule::writeBim(const string & file, vector<Locus*> & locus)
{
    ofstream BIT;

    if (debug == 2) cout << "Writing map (extended format) information to [ " + file + " ] \n";

    BIT.open(file.c_str(), ios::out);
    if (!BIT) {
	cout << "Error: Cannot write to output file " << (file) << "\n";
	gracefulExit(1);
    }

    int nlocus = locus.size();

    for (int l = 0; l < nlocus; l++) {

	Locus *p = locus[l];

	if (p->allele1 == 0) {
	    if (p->allele2 != '0') p->allele1 = '0';
	    else p->allele1 = 'X';
	}

	if (p->allele2 == 0) p->allele2 = '0';

	BIT << p->chr << "\t" << p->name << "\t" << p->pos << "\t" << p->bp << "\t" << p->allele1
		<< "\t" << p->allele2 << "\n";
    }

    BIT.clear();
    BIT.close();
}

/**
 * Write a .bed plink binary file
 */

void plinkModule::writeBed(const string & head, vector<Individual*> &persons,
	vector<Locus*> &locus)
{

    if (debug == 2) cout << ("Writing genotype bitfile to [ " + head + ".bed ] \n");

    ofstream BIT;
    BIT.open((head + ".bed").c_str(), ios::out | ios::binary);
    if (!BIT) {
	cout << "Cannot write to output file " << (head + ".bed") << "\n";
	gracefulExit(1);
    }

    int npersons = persons.size();
    int nlocus = locus.size();

    // for (int p = 0; p < npersons; p++)  cout << persons[p]->id << " "; cout << "\n";


    // BED files begin with characters 6c 1b
    // 0 specifies individual major order
    // 1 specifies SNP major order
    char filehead[3] = { 0x6c, 0x1b, 0 };

    if (out_SNP_major == 1) filehead[2] = 1;

    BIT.write(filehead, 3);

    char *data; // temporary buffer
    int byte; // packed 4 genotypes into one byte
    int data_p, pos;

    if (out_SNP_major == 0) // Write in individual major mode
    {
	data = new char[4 + nlocus / 4];

	for (int p = 0; p < npersons; p++) {
	    Individual *person = persons[p];

	    byte = 0;
	    data_p = 0;
	    pos = 0;

	    for (int snp = 0; snp < nlocus; snp++) {
		Locus *lp = locus[snp];

		//byte |= (person->one[lp->id] | (2 * person->two[lp->id])) << (2 * pos);
		byte |= person->getGeno(lp->id) << (2 * pos);

		//cout << person->one[snp] << "." << person->two[snp] << "  " << byte << " ";
		pos++;
		if (pos == 4) {
		    pos = 0;
		    data[data_p++] = byte;
		    byte = 0;
		}
	    }

	    if (pos != 0) data[data_p++] = byte;

	    //for(int k=0;k<data_p;k++) printf("%d [%x]\n",k,(int)data[k]);
	    //printf("Write %d bytes\n", data_p);
	    BIT.write(data, data_p);
	}

    }
    else // Write in SNP Major Mode
    {
	data = new char[4 + npersons / 4];

	for (int snp = 0; snp < nlocus; snp++) { // iterate for every SNP
	    Locus *lp = locus[snp];

	    byte = 0;
	    data_p = 0;
	    pos = 0;

	    for (int p = 0; p < npersons; p++) { // for every individual
		//byte |= (persons[p]->one[lp->id] | (2 * persons[p]->two[lp->id])) << (2 * pos);
		byte |= persons[p]->getGeno(lp->id) << (2 * pos);

		//cout << person->one[snp] << "." << person->two[snp] << "  " << byte << " ";
		pos++;
		if (pos == 4) {
		    pos = 0;
		    data[data_p++] = byte;
		    byte = 0;
		}
	    }

	    if (pos != 0) data[data_p++] = byte;

	    //printf("Write %d bytes\n", data_p);
	    BIT.write(data, data_p);
	}
    }

    delete[] data;
    BIT.close();
}

/**
 * plink::writeBinFiles
 * Write a set of binary output files
 *
 *  Options:
 *		head (string) : Prefix for output files
 *		persons (vector<Individual*>) : Individuals to write
 *		locus (vector<Locus*>) : Which SNPs to write
 *		newfids (int) : If equal to 1, then generate new individual iid
 *			to avoid duplicates generated by bootstrap
 *
 */

void plinkModule::writeBinFiles(const string &head, vector<Individual*> &persons,
	vector<Locus*> &locus, int newfids)
{
    int npersons = persons.size();

    if (debug >= 2) cout << "Writing " << persons.size() << "*" << locus.size() << " data to "
	    << head << " newfids=" << newfids << "\n";

    writeBim(head + ".bim", locus);

    // Write the fam file

    ofstream BIT((head + ".fam").c_str(), ios::out);
    if (!BIT) {
	cout << "Cannot write to output file " << (head + ".fam") << "\n";
	gracefulExit(1);
    }

    // For each individual
    for (int i = 0; i < npersons; i++) {
	Individual * person = persons[i];

	if (newfids)
	    BIT << person->fid << " " << "i" << i << " ";
	else
	    BIT << person->fid << " " << person->iid << " ";

	BIT << person->pat << " " << person->mat << " " << person->sexcode << " ";

	BIT << person->phenotype << "\n";

    }

    BIT.clear();
    BIT.close();

    writeBed(head, persons, locus);

    if (haveCovariates == true) {
	writeCovariateFile(head + ".cov", persons, newfids);
    }

}


void plinkModule::deleteMissingPersons()
{
    int ndel = 0;
    vector<Individual*> persons2;


    for(int i=0; i<npersons; i++)
	if(persons[i]->phenotype == NA)
	    { delete persons[i]; ndel++; }
	else
	    persons2.push_back(persons[i]);

    if(ndel > 0)
	cout << "Warning: Deleting " << ndel << " individuals with missing phenotype.\n\n";

    persons = persons2;
    npersons = persons2.size();
}

void plinkModule::readCovariateFile(string in)
{
    cout << "Reading covariate file " << in << "\n";

    map<string, int> covmap;
    map<string, int>::iterator covmap_i;

    string s;
    char fid[100], iid[100];

    ifstream cov(in.c_str());
    if(!cov) {
	cout << "Error: Cannot open covariates file " << in << "\n";
	gracefulExit(1);
    }

    for(int i=0; i<npersons; i++) covmap[ persons[i]->fid + "," + persons[i]->iid ] = i;

    // The first line might be a header row.
    int i, n;

    int line = 1;

    int numcov = 0;

    while (getline(cov, s)) {

	istringstream a(s);
	string rest;

	a >> fid >> iid;
	getline(a, rest);

	covmap_i = covmap.find((string)fid + "," + iid);

	// Might be a header line
        if(line == 1 && covmap_i == covmap.end()) {
            cout << "Covariates header line: " << s << "\n";
            par::covariate_header = s;
            par::covariate_names = split(rest);
            line++;
            continue;
        }

	// cout << "Found: " << fid << " " << iid << " " << covmap_i->second << "\n";

	if(covmap_i == covmap.end()) {
	    cout << "Warning: Individual (" << fid << "," << iid << ") present in covariates file does not appear in genotype data." << "\n";
	    line++;
	    continue;
	}

	vector<string> covs = split(rest);

	numcov = covs.size();

	for(int i=0; i < covs.size(); i++ ) {
	    double pheno = getPhenotype(covs[i]);
	    persons[covmap_i->second]->covariates.push_back ( pheno );
	}

	line++;
    }

    cout << "Number of covariates: " << numcov << "\n";

    for(i = 0; i < npersons; i++) if(persons[i]->covariates.size() == 0) {
	cout << "Warning: Cannot find covariates for individual "
		+ persons[i]->fid +  "," + persons[i]->iid + "\n";

	for(int j = 0; j < numcov; j++) persons[i]->covariates.push_back(NA);
    }

    ofstream out("model.txt");

    for(i = 0; i < npersons; i++) {
	out << persons[i]->fid << " " << persons[i]->iid << " " << persons[i]->covariates.size() << "\t";
	vector<double> c = persons[i]->covariates;
	out << toString(persons[i]->phenotype) << " ~ G ";
	for(int j=0;j<c.size();j++) out << toString(c[j]) << " ";
	out << "\n";
    }
}

void plinkModule::writeCovariateFile(string outFileName, const vector<Individual*> &persons, int newfids)
{
    ofstream cov(outFileName.c_str());

    if(par::covariate_header != "") cov << par::covariate_header << "\n";

    int npersons = persons.size();

    for (int i = 0; i < npersons; i++) {
	Individual * person = persons[i];

	if (newfids)
	    cov << person->fid << " " << "i" << i << " ";
	else
	    cov << person->fid << " " << person->iid << " ";

	for(int j = 0; j < person->covariates.size(); j ++)
	    if( person->covariates[j] == NA )
		cov << "-9 ";
	    else
		cov << (person->covariates[j]) << " ";

	cov << "\n";
    }
}

void plinkModule::readAltPhenotype(string fileName)
{
    cout << "Reading alternative phenotype from file " << fileName << "\n";

    map<string, int> personmap;
    map<string, int>::iterator personmap_i;

    string s;
    char fid[100], iid[100];

    ifstream phenofile(fileName.c_str());
    if(!phenofile) {
	cout << "Error: Cannot read file " << fileName << "\n";
	gracefulExit(1);
    }

    for(int i=0;i<npersons;i++) personmap[ persons[i]->fid + "," + persons[i]->iid ] = i;

    for(int i=0;i<npersons;i++) persons[i]->phenotype = NA; // missing

    // The first line might be a header row.
    int i, n;

    int line = 0;
    int notfound = 0;

    while (getline(phenofile, s)) {
	line ++;
	string pheno;
	istringstream a(s);
	a >> fid >> iid >> pheno;

	personmap_i = personmap.find((string)fid + "," + iid);

	if(line == 1) // Might be a header line
	    if(personmap_i == personmap.end()) continue;

	if(personmap_i == personmap.end())
	    {
		notfound++;
		continue;
	    }

	double ph = getPhenotype(pheno);
	persons[personmap_i->second]->phenotype = ph;
    }

    notfound = 0;
    for(int i=0;i<npersons;i++) if(persons[i]->phenotype == NA) notfound++;

    cout << "Read alternative phenotype for " << (npersons-notfound) << " individuals\n";
    if(notfound > 0) cout << "Warning: " << notfound << " individuals with no phenotype.\n";
}

/* Check if phenotype is missing and convert it to double */

double plinkModule::getPhenotype(string pheno)
{
    if(pheno == par::missing_phenotype) return NA;

    double ph = NA;
    int nread = sscanf(pheno.c_str(), "%lf", &ph);
    if(nread != 1) ph = NA;

    return ph;
}

/* Using plink R-plugin to compute association and p-values */

void plinkModule::runRplugin(const string & binfile,
	vector<LogisticOutput> &results, vector<Locus*> &locus, string pluginFile)
{
    // Run Plink

    string cmdline = "--bfile \"" + binfile + "\" " + "--R " + pluginFile + " --out \"" + binfile + "\"";

    if (haveCovariates == true) {
	cmdline += " --covar " + binfile + ".cov " + par::covariate_command_line;
    }

    if(locus.size() == 0) {
	results.clear();
	return;
    }

    runPLink(cmdline);

    // Read plink output file

    ifstream in;
    in.open((binfile + ".auto.R").c_str());

    //system(("plink --bfile " + binfile + " --model --silent").c_str());
    //system(((string)"cat plink.model").c_str());
    //system(("cat " + binfile + ".assoc.logistic").c_str());

    if (!in) {
	cout << "Cannot open plink output file: " + (binfile + ".assoc.logistic") + "\n";
	gracefulExit(1);
    }

    results.clear();
    int nlocus = locus.size();

    char s[10000]; // Line currently read from input file

    int chr, bp;
    char snp[512], allele[16], test[32], odds_s[100], p_s[100];

    double odds, p;

    int pos = 0;
    int line = 1;

    while (in.getline(s, 10000)) {
	line++;

	int n = sscanf(s, "%d %s %d %s %s %s", &chr, snp, &bp, allele, odds_s, p_s);

	if (n != 6) {
	    cout << "Error: Cannot read enough items from output file " << (binfile
		    + ".auto.R") << " line " << line << "\n";
	    cout << "Offending line is: " << s << "\n\n";
	    gracefulExit(1);
	}

	if(strcmp(p_s,"NA")!=0) if(toDouble(p_s)<0 || toDouble(p_s)>1) {
	    cout << "Error: p-value from R-plugin is not between 0,1 : " << p_s << "\n";
	    gracefulExit(1);
	}


	if(debug) printf("%s %s %s %s\n",snp,test,odds_s,p_s);

	if (pos >= nlocus) {
	    cout << "Error: Additional SNPs in output file: " << (binfile + ".auto.R")
		    << "!\n";
	    gracefulExit(1);
	}

	if (strcmp(locus[pos]->name.c_str(), snp) != 0) {
	    cout << "Error: SNP's out of order on output file " << (binfile + ".auto.R")
		    << "    line:" << pos << "  expected:" << locus[pos]->name << "  got: " << snp
		    << "\n";

	    cout << "br2 input files should be sorted according to chr/bp. "
	            "Please recode input file with plink (plink --recode) and run br2 again.\n";

	    gracefulExit(1);
	}

	odds = toDouble(odds_s);
	p = toDouble(p_s);

	if(strcmp(odds_s, "NA") == 0) { odds = NA; p = NA; }
	if(strcmp(p_s, "NA") == 0) { odds = NA; p = NA; }

	LogisticOutput r(locus[pos], allele[0], exp(odds), p);
	results.push_back(r);

	pos++;
    }

    if(results.size() != locus.size()) {
	cout << "Error: Cannot read all test results from plink (got "
	     << results.size() << " results while expecting " << locus.size() << " )\n";

	gracefulExit(1);
    }
}


